# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import re
from subprocess import check_output, check_call, call
import time

from path import Path
from jujuresources import resource_path
from jujuresources import fetch_resources
from jujuresources import verify_resources

from charmhelpers.core import host
from charmhelpers.core import hookenv
from charmhelpers.core import charmdata
from charmhelpers import fetch
from charmhelpers.contrib.bigdata import utils


class HadoopBase(object):
    VENDOR = 'apache'
    HADOOP_VERSION = '2.4.1'
    JAVA_VERSION = '7'
    SUPPORTED_ARCHS = ['x86_64', 'ppc64le']
    PACKAGES = [
        'libsnappy1',
        'libsnappy-dev',
        'openssl',
        'liblzo2-2',
    ]
    DIRS = {
        'demo': Path('/home/ubuntu'),
        'hosts': Path('/etc/hosts'),
        'hadoop': Path('/usr/lib/hadoop'),
        'hadoop_installed': Path('/usr/lib/hadoop/hadoop-%s' % HADOOP_VERSION),
        'hadoop_conf': Path('/etc/hadoop/conf.juju'),
        'hadoop_real_conf': Path('/etc/hadoop/conf'),
        'hadoop_tmp': Path('/tmp/hadoop'),
        'yarn_tmp': Path('/tmp/hadoop-yarn'),
        'mapred_log': Path('/var/log/hadoop/mapred'),
        'mapred_run': Path('/var/run/hadoop/mapred'),
    }

    def __init__(self):
        self.cpu_arch = check_output(['uname', '-p']).strip()
        if self.cpu_arch not in self.SUPPORTED_ARCHS:
            raise NotImplementedError('Unsupported archecture: %s' % self.cpu_arch)
        self.spec = '; '.join(sorted('%s=%s' % (k, v) for k, v in {
            'vendor': self.VENDOR,
            'hadoop': self.HADOOP_VERSION,
            'java': self.JAVA_VERSION,
            'arch': self.cpu_arch,
        }.iteritems()))

    def verify_conditional_resources(self):
        conditional_resources = ['hadoop-%s' % self.cpu_arch]
        if self.cpu_arch == 'ppc64le':
            conditional_resources.append('ibm-java-installer')
        if verify_resources(conditional_resources):
            return True
        else:
            fetch_resources(conditional_resources)
            return verify_resources(conditional_resources)

    def is_installed(self):
        return charmdata.kv.get('hadoop.base.installed')

    def install(self, force=False):
        if not force and self.is_installed():
            return
        self.configure_hosts_file()
        self.manage_users()
        self.manage_dirs()
        self.install_base_packages()
        self.setup_hadoop_config()
        self.configure_hadoop()
        self.config_for_hue()
        charmdata.kv.set('hadoop.base.installed', True)

    def configure_hosts_file(self):
        """
        Add the unit's private-address to /etc/hosts to ensure that Java
        can resolve the hostname of the server to its real IP address.
        """
        private_address = hookenv.unit_get('private-address')
        hostname = check_output(['hostname']).strip()
        hostfqdn = check_output(['hostname', '-f']).strip()
        etc_hosts = Path('/etc/hosts')
        hosts = etc_hosts.lines()
        line = '%s %s %s' % (private_address, hostfqdn, hostname)
        IP_pat = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}'
        if not re.match(IP_pat, private_address):
            line = '# %s  # private-address did not return an IP' % line
        if hosts[0] != line:
            hosts.insert(0, line)
            etc_hosts.write_lines(hosts)

    def manage_users(self):
        host.add_group('hadoop')
        host.add_group('mapred')
        host.add_user_to_group('ubuntu', 'hadoop')
        host.add_user_to_group('ubuntu', 'mapred')
        host.adduser('hdfs', group='hadoop')
        host.adduser('mapred', group='hadoop', groups=['mapred'])
        host.adduser('yarn', group='hadoop')
        host.adduser('hue', group='hadoop')

    def manage_dirs(self):
        config = hookenv.config()
        host.mkdir(config['yarn_log_dir'], owner='yarn', group='hadoop', perms=0o755)
        host.mkdir(config['hdfs_log_dir'], owner='hdfs', group='hadoop', perms=0o755)
        host.mkdir(self.DIRS['mapred_log'], owner='mapred', group='hadoop', perms=0o755)
        host.mkdir(self.DIRS['mapred_run'], owner='mapred', group='hadoop', perms=0o755)
        host.mkdir(self.DIRS['hadoop'], perms=0o777)
        host.mkdir(self.DIRS['hadoop_tmp'], perms=0o777)
        host.mkdir(self.DIRS['yarn_tmp'], perms=0o777)
        datadir = Path(config['hadoop_dir_base'])
        host.mkdir(datadir / 'cache', owner='hdfs', group='hadoop', perms=0o1775)
        host.mkdir(datadir / 'cache/hadoop', owner='hdfs', group='hadoop', perms=0o775)
        host.mkdir(self.DIRS['hadoop_installed'] / 'logs', owner='mapred', group='hadoop', perms=0o1777)

    def install_base_packages(self):
        with utils.disable_firewall():
            fetch.apt_update()
            fetch.apt_install(self.PACKAGES)
            self.install_jdk()
            self.install_hadoop()

    def find_jdk(self):
        if self.cpu_arch == 'x86_64':
            output = check_output(['find', '/usr/', '-name', 'java-%s-openjdk-*' % self.JAVA_VERSION])
        elif self.cpu_arch == 'ppc64le':
            output = check_output(['find', '/opt/', '-name', 'java-ppc64le-*'])
        if output.strip():
            return Path(output.split('\n')[0])
        else:
            return None

    def install_jdk(self):
        if self.find_jdk():
            return
        if self.cpu_arch == 'ppc64le':
            java_installer = Path(resource_path('ibm-java-installer'))
            java_installer.chmod('+x')
            check_call([java_installer, '-i', 'silent'])
        else:
            fetch.apt_install(['openjdk-%s-jdk' % self.JAVA_VERSION])

    def install_hadoop(self):
        hadoop_pkg = resource_path('hadoop-%s' % self.cpu_arch)
        check_call(['tar', '-xzf', hadoop_pkg, '-C', self.DIRS['hadoop']])

    def setup_hadoop_config(self):
        # copy default config into alternate dir
        conf_dir = self.DIRS['hadoop_installed'] / 'etc/hadoop'
        self.DIRS['hadoop_conf'].rmtree_p()
        conf_dir.copytree(self.DIRS['hadoop_conf'])
        (self.DIRS['hadoop_conf'] / 'slaves').remove_p()
        mapred_site = self.DIRS['hadoop_conf'] / 'mapred-site.xml'
        if not mapred_site.exists():
            (self.DIRS['hadoop_conf'] / 'mapred-site.xml.template').copy(mapred_site)
        # setup to use alternates
        if call(['update-alternatives', '--list', 'hadoop-conf']) != 0:
            check_call(['update-alternatives', '--install',
                        self.DIRS['hadoop_real_conf'], 'hadoop-conf',
                        self.DIRS['hadoop_conf'], '50'])

    def configure_hadoop(self):
        config = hookenv.config()
        jdk_path = self.find_jdk()
        if not jdk_path:
            raise ValueError('Unable to find JDK')

        jdk_bin = jdk_path / 'bin'
        hadoop_bin = self.DIRS['hadoop_installed'] / 'bin'
        hadoop_sbin = self.DIRS['hadoop_installed'] / 'bin'
        with utils.environment_edit_in_place('/etc/environment') as env:
            env['JAVA_HOME'] = jdk_path
            if jdk_bin not in env['PATH']:
                env['PATH'] = ':'.join([env['PATH'], jdk_bin])
            if hadoop_bin not in env['PATH']:
                env['PATH'] = ':'.join([env['PATH'], hadoop_bin])
            if hadoop_sbin not in env['PATH']:
                env['PATH'] = ':'.join([env['PATH'], hadoop_sbin])
            env['HADOOP_LIBEXEC_DIR'] = self.DIRS['hadoop_installed'] / 'libexec'
            env['HADOOP_INSTALL'] = self.DIRS['hadoop_installed']
            env['HADOOP_HOME'] = self.DIRS['hadoop_installed']
            env['HADOOP_COMMON_HOME'] = self.DIRS['hadoop_installed']
            env['HADOOP_HDFS_HOME'] = self.DIRS['hadoop_installed']
            env['HADOOP_MAPRED_HOME'] = self.DIRS['hadoop_installed']
            env['HADOOP_YARN_HOME'] = self.DIRS['hadoop_installed']
            env['YARN_HOME'] = self.DIRS['hadoop_installed']
            env['HADOOP_CONF_DIR'] = self.DIRS['hadoop_conf']
            env['YARN_CONF_DIR'] = self.DIRS['hadoop_conf']
            env['YARN_LOG_DIR'] = config['yarn_log_dir']
            env['HDFS_LOG_DIR'] = config['hdfs_log_dir']
            env['HADOOP_LOG_DIR'] = config['hdfs_log_dir']  # for hadoop 2.2.0 only
            env['MAPRED_LOG_DIR'] = '/var/log/hadoop/mapred'  # should be moved to config, but could
            env['MAPRED_PID_DIR'] = '/var/run/hadoop/mapred'  # be destructive for mapreduce operation

        hadoop_env = self.DIRS['hadoop_conf'] / 'hadoop-env.sh'
        utils.re_edit_in_place(hadoop_env, {
            r'export JAVA_HOME *=.*': 'export JAVA_HOME=%s' % jdk_path,
        })

        core_site = self.DIRS['hadoop_conf'] / 'core-site.xml'
        with utils.xmlpropmap_edit_in_place(core_site) as props:
            props["io.file.buffer.size"] = config["io_file_buffer_size"]

    def config_for_hue(self):
        # Set hue-specific properties in our hdfs|core-site.xml files
        hdfs_site = self.DIRS['hadoop_conf'] / 'hdfs-site.xml'
        with utils.xmlpropmap_edit_in_place(hdfs_site) as props:
            props['dfs.webhdfs.enabled'] = "true"
        core_site = self.DIRS['hadoop_conf'] / 'core-site.xml'
        with utils.xmlpropmap_edit_in_place(core_site) as props:
            props['hadoop.proxyuser.hue.hosts'] = "*"
            props['hadoop.proxyuser.hue.groups'] = "*"

    def run(self, user, command, *args):
        """
        Run a Hadoop command as the `hdfs` user.

        :param str command: Command to run, prefixed with `bin/` or `sbin/`
        :param list args: Additional args to pass to the command
        """
        parts = [self.DIRS['hadoop_installed'] / command] + list(args)
        quoted = ' '.join("'%s'" % p for p in parts)
        check_call(['su', user, '-c', quoted])


class HDFS(object):
    master_ports = [8020, 50070]
    slave_ports = [50070, 50010, 50020, 50075]

    def __init__(self, hadoop_base):
        self.hadoop_base = hadoop_base

    def stop_master(self):
        self._hadoop_daemon('stop', 'namenode')

    def start_master(self):
        self._hadoop_daemon('start', 'namenode')
        # Some hadoop processes take a bit of time to start
        # we need to let them get to a point where they are
        # ready to accept connections - increase the value for hadoop 2.4.1
        time.sleep(30)

    def restart_master(self):
        self.stop_master()
        self.start_master()

    def stop_slave(self):
        self._hadoop_daemon('stop', 'datanode')

    def start_slave(self):
        self._hadoop_daemon('start', 'datanode')

    def restart_slave(self):
        self.stop_slave()
        self.start_slave()

    def configure_master(self):
        if not charmdata.kv.get('hdfs.master.configured'):
            self.configure_hdfs(master=True)
            self.format_namenode()
            self.create_hdfs_dirs()
            charmdata.kv.set('hdfs.master.configured', True)

    def configure_slave(self):
        if not charmdata.kv.get('hdfs.slave.configured'):
            self.configure_hdfs(master=False)
            self.start_slave()
            charmdata.kv.set('hdfs.slave.configured', True)

    def configure_client(self):
        self.configure_hdfs(master=False)

    def configure_hdfs(self, master):
        config = hookenv.config()
        if master:
            host = hookenv.unit_get('private-address')
        else:
            hdfs_master = charmdata.kv.get('relations.ready')['hdfs-master']
            host = hdfs_master.values()[0]['private-address']
        core_site = self.hadoop_base.DIRS['hadoop_conf'] / 'core-site.xml'
        with utils.xmlpropmap_edit_in_place(core_site) as props:
            props['fs.defaultFS'] = "hdfs://{host}:8020".format(host=host)
        hdfs_site = self.hadoop_base.DIRS['hadoop_conf'] / 'hdfs-site.xml'
        with utils.xmlpropmap_edit_in_place(hdfs_site) as props:
            props['dfs.namenode.name.dir'] = Path(config['hadoop_dir_base']) / 'cache/hadoop/dfs/name'
            props['dfs.namenode.handler.count'] = config['dfs_namenode_handler_count']
            props['dfs.datanode.max.transfer.threads'] = config['dfs_datanode_max_xcievers']
            props['dfs.datanode.data.dir'] = Path(config['hadoop_dir_base']) / 'cache/hadoop/dfs/name'
            props['dfs.permissions'] = 'false'  # TODO - secure this hadoop installation!
            props['dfs.replication'] = config['dfs_replication']
            props['dfs.heartbeat.interval'] = config['dfs_heartbeat_interval']
            props['dfs.namenode.heartbeat.recheck-interval'] = config['dfs_namenode_heartbeat_recheck_interval']

    def format_namenode(self):
        self.stop_master()
        self._hdfs('namenode', '-format', '-force')
        self.start_master()

    def create_hdfs_dirs(self):
        self._hdfs('dfs', '-mkdir', '-p', '/tmp/hadoop/mapred/staging')
        self._hdfs('dfs', '-chmod', '-R', '1777', '/tmp/hadoop/mapred/staging')
        self._hdfs('dfs', '-mkdir', '-p', '/tmp/hadoop-yarn/staging')
        self._hdfs('dfs', '-chmod', '-R', '1777', '/tmp/hadoop-yarn')
        self._hdfs('dfs', '-mkdir', '-p', '/user/ubuntu')
        self._hdfs('dfs', '-chown', '-R', 'ubuntu', '/user/ubuntu')
        self._hdfs('dfs', '-mkdir', '-p', '/user/hue')
        self._hdfs('dfs', '-chown', '-R', 'hue', '/user/hue')
        # for JobHistory
        self._hdfs('dfs', '-mkdir', '-p', '/mr-history/tmp')
        self._hdfs('dfs', '-chmod', '-R', '1777', '/mr-history/tmp')
        self._hdfs('dfs', '-mkdir', '-p', '/mr-history/done')
        self._hdfs('dfs', '-chmod', '-R', '1777', '/mr-history/done')
        self._hdfs('dfs', '-chown', '-R', 'mapred:hdfs', '/mr-history')
        self._hdfs('dfs', '-mkdir', '-p', '/app-logs')
        self._hdfs('dfs', '-chmod', '-R', '1777', '/app-logs')
        self._hdfs('dfs', '-chown', 'yarn', '/app-logs')

    def register_slaves(self):
        slaves = charmdata.kv.get('relations.ready', {}).get('hdfs-master-slave', {})
        slaves_file = self.hadoop_base.DIRS['hadoop_conf'] / 'slaves'
        slaves_file.write_lines(
            [
                '# DO NOT EDIT',
                '# This file is automatically managed by Juju',
            ] + [
                slave['private-address'] for slave in slaves.values()
            ]
        )
        slaves_file.chown('ubuntu', 'hadoop')

    def _hadoop_daemon(self, command, service):
        self.hadoop_base.run('hdfs', 'sbin/hadoop-daemon.sh',
                             '--config', self.hadoop_base.DIRS['hadoop_conf'],
                             command, service)

    def _hdfs(self, command, *args):
        self.hadoop_base.run('hdfs', 'bin/hdfs', command, *args)


class YARN(object):
    master_ports = [50070, 8030, 8031, 8032, 8045, 8050, 8088, 19888, 10020]
    slave_ports = [50070, 8030, 8031, 8032, 8045, 8050, 8088, 19888, 10020]

    def __init__(self, hadoop_base):
        self.hadoop_base = hadoop_base

    def stop_master(self):
        self._yarn_daemon('stop', 'resourcemanager')
        self._jobhistory_daemon('stop', 'historyserver')

    def start_master(self):
        self._yarn_daemon('start', 'resourcemanager')
        self._jobhistory_daemon('start', 'historyserver')

    def restart_master(self):
        self.stop_master()
        self.start_master()

    def stop_slave(self):
        self._yarn_daemon('stop', 'nodemanager')

    def start_slave(self):
        self._yarn_daemon('start', 'nodemanager')

    def restart_slave(self):
        self.stop_slave()
        self.start_slave()

    def configure_master(self):
        if not charmdata.kv.get('yarn.master.configured'):
            self.configure_yarn(master=True)
            charmdata.kv.set('yarn.master.configured', True)
            self.start_master()

    def configure_slave(self):
        if not charmdata.kv.get('yarn.slave.configured'):
            self.configure_yarn(master=False)
            self.start_slave()
            charmdata.kv.set('yarn.slave.configured', True)

    def configure_client(self):
        self.configure_yarn(master=False)
        self.install_demo()

    def configure_yarn(self, master):
        config = hookenv.config()
        if master:
            host = hookenv.unit_get('private-address')
        else:
            yarn_master = charmdata.kv.get('relations.ready')['yarn-master']
            host = yarn_master.values()[0]['private-address']
        yarn_site = self.hadoop_base.DIRS['hadoop_conf'] / 'yarn-site.xml'
        with utils.xmlpropmap_edit_in_place(yarn_site) as props:
            props["yarn.resourcemanager.resource-tracker.address"] = "{host}:8031".format(host=host)
            props["yarn.resourcemanager.scheduler.address"] = "{host}:8030".format(host=host)
            props["yarn.resourcemanager.address"] = "{host}:8032".format(host=host)
            props["yarn.log.server.url"] = "{host}:19888/jobhistory/logs/".format(host=host)
            props["yarn.resourcemanager.webapp.address"] = "{host}:8088".format(host=host)
            props["yarn.nodemanager.aux-services"] = config["yarn_nodemanager_aux-services"]
            props["yarn.nodemanager.aux-services.mapreduce.shuffle.class"] = \
                config["yarn_nodemanager_aux-services_mapreduce_shuffle_class"]
            if master:
                props["yarn.nodemanager.address"] = "{host}:8050".format(host=host)
                props["yarn.nodemanager.localizer.address"] = "{host}:8045".format(host=host)
        mapred_site = self.hadoop_base.DIRS['hadoop_conf'] / 'mapred-site.xml'
        with utils.xmlpropmap_edit_in_place(mapred_site) as props:
            props["mapreduce.jobhistory.address"] = "{host}:10020".format(host=host)
            props["mapreduce.jobhistory.webapp.address"] = "{host}:19888".format(host=host)
            props["mapreduce.framework.name"] = config["mapreduce_framework_name"]
            props["mapreduce.reduce.shuffle.parallelcopies"] = config["mapreduce_reduce_shuffle_parallelcopies"]
            props["mapred.child.java.opts"] = config["mapred_child_java_opts"]
            props["mapreduce.task.io.sort.factor"] = config["mapreduce_task_io_sort_factor"]
            props["mapreduce.task.io.sort.mb"] = config["mapreduce_task_io_sort_mb"]
            props["mapred.job.tracker.handler.count"] = config["mapred_job_tracker_handler_count"]
            props["tasktracker.http.threads"] = config["tasktracker_http_threads"]

    def install_demo(self):
        # Copy our demo (TeraSort) to the target location and set mode/owner
        demo_source = 'scripts/terasort.sh'
        demo_target = self.hadoop_base.DIRS['demo'] / 'terasort.sh'

        Path(demo_source).copy(demo_target)
        Path(demo_target).chmod(0o755)
        Path(demo_target).chown('ubuntu', 'hadoop')

    def _yarn_daemon(self, command, service):
        self.hadoop_base.run('yarn', 'sbin/yarn-daemon.sh',
                             '--config', self.hadoop_base.DIRS['hadoop_conf'],
                             command, service)

    def _jobhistory_daemon(self, command, service):
        # TODO refactor job history to separate class
        self.hadoop_base.run('mapred', 'sbin/mr-jobhistory-daemon.sh',
                             '--config', self.hadoop_base.DIRS['hadoop_conf'],
                             command, service)
