# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

from subprocess import check_call

from path import Path

from charmhelpers.core import host
from charmhelpers.core.ch_framework.helpers import Relation


class SpecMatchingRelation(Relation):
    """
    Relation base class that validates that a version and environment
    between two related charms match, to prevent interoperability issues.

    This class adds a `spec` key to the `required_keys` and populates it
    in :meth:`provide`.  The `spec` value must be passed in to :meth:`__init__`.

    The `spec` string should describe all aspects of the charm's environment
    or configuration that might affect its interoperability with the remote
    charm.

    An example spec string might be: `"arch=x86_64; vendor=apache; version=2.4"`
    """
    def __init__(self, spec, *args, **kwargs):
        """
        Create a new relation handler instance.

        :param str spec: Spec string that should capture version or environment
            particulars which can cause issues if mismatched.
        """
        super(SpecMatchingRelation, self).__init__(*args, **kwargs)
        if not spec:
            raise ValueError('Empty spec')
        self.spec = spec
        self.required_keys = ['spec'] + self.required_keys

    def provide(self, remote_service, ready):
        """
        Provide the `spec` data to the remote service.

        Subclasses *must* either delegate to this method (e.g., via `super()`)
        or include `'spec': self.spec` in the provided data themselves.
        """
        return {
            'spec': self.spec,
        }

    def is_ready(self):
        """
        Validate the `spec` data from the connected units to ensure that
        it matches the local `spec` exactly.
        """
        if not super(SpecMatchingRelation, self).is_ready():
            return False
        for unit, data in self.filtered_data().iteritems():
            if data.get('spec') != self.spec:
                # TODO XXX Once extended status reporting is available,
                #          we should use that instead of erroring.
                raise ValueError(
                    'Version or environment mismatch with related unit %s: '
                    '%r != %r' % (unit, data.get('spec'), self.spec))
        return True


class HDFSMaster(SpecMatchingRelation):
    relation_name = 'hdfs-master'
    required_keys = ['private-address']
    optional_keys = ['ready']

    def __init__(self, spec, relation=None, *args, **kwargs):
        super(HDFSMaster, self).__init__(spec, *args, **kwargs)
        if relation:
            self.relation_name = 'hdfs-master-{}'.format(relation)

    def provide(self, remote_service, ready):
        data = super(HDFSMaster, self).provide(remote_service, ready)
        if ready:
            data['ready'] = 'true'
        return data


class HDFSSlave(Relation):
    relation_name = 'hdfs-master-slave'
    required_keys = ['private-address']


class YARNMaster(SpecMatchingRelation):
    relation_name = 'yarn-master'
    required_keys = ['private-address', 'ssh-key']

    def get_ssh_key(self):
        sshdir = Path('/home/ubuntu/.ssh')
        keyfile = sshdir / 'id_rsa'
        pubfile = sshdir / 'id_rsa.pub'
        if not pubfile.exists():
            (sshdir / 'config').write_lines([
                'Host *',
                '    StrictHostKeyChecking no'
            ], append=True)
            check_call(['ssh-keygen', '-t', 'rsa', '-P', '', '-f', keyfile])
            host.chownr(sshdir, 'ubuntu', 'hadoop')
        return pubfile.text()

    def provide(self, remote_service, ready):
        data = super(YARNMaster, self).provide(remote_service, ready)
        data.update({
            'ssh-key': self.get_ssh_key(),
        })
        if ready:
            data['ready'] = 'true'
        return data

    def install_ssh_keys(self):
        ssh_keys = self.filtered_relation_data().values()
        Path('/home/ubuntu/.ssh/authorized_keys').write_lines(ssh_keys, append=True)
